import pytest

from winrm.exceptions import (
    WinRMOperationTimeoutError,
    WinRMTransportError,
    WSManFaultError,
)


def raise_exc(exc: Exception) -> None:
    raise exc


def test_wsman_fault_must_understand(protocol_fake):
    xml_text = r"""<s:Envelope xml:lang="en-US"
    xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing"
    xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer"
    xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing"
    xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
    xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"
    xmlns:rsp="http://schemas.microsoft.com/wbem/wsman/1/windows/shell"
    xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <a:Action>http://schemas.xmlsoap.org/ws/2004/08/addressing/fault</a:Action>
        <a:MessageID>uuid:4DB571F9-F8DE-48FD-872C-2AF08D996249</a:MessageID>
        <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
        <a:RelatesTo>uuid:eaa98952-3188-458f-b265-b03ace115f20</a:RelatesTo>
        <s:NotUnderstood qname="wsman:ResourceUri"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" />
    </s:Header>
    <s:Body>
        <s:Fault>
            <s:Code>
                <s:Value>s:MustUnderstand</s:Value>
            </s:Code>
            <s:Reason>
                <s:Text xml:lang=""> Test reason. </s:Text>
            </s:Reason>
        </s:Fault>
    </s:Body>
</s:Envelope>"""

    protocol_fake.transport.send_message = lambda m: raise_exc(WinRMTransportError("http", 500, xml_text))

    with pytest.raises(WSManFaultError, match="Test reason\\.") as exc:
        protocol_fake.open_shell()

    assert isinstance(exc.value, WSManFaultError)
    assert exc.value.code == 500
    assert exc.value.response == xml_text
    assert exc.value.fault_code == "s:MustUnderstand"
    assert exc.value.fault_subcode is None
    assert exc.value.wsman_fault_code is None
    assert exc.value.wmierror_code is None


def test_wsman_fault_no_reason(protocol_fake):
    xml_text = r"""<s:Envelope xml:lang="en-US"
    xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing"
    xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer"
    xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing"
    xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
    xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"
    xmlns:rsp="http://schemas.microsoft.com/wbem/wsman/1/windows/shell"
    xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <a:Action>http://schemas.xmlsoap.org/ws/2004/08/addressing/fault</a:Action>
        <a:MessageID>uuid:4DB571F9-F8DE-48FD-872C-2AF08D996249</a:MessageID>
        <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
        <a:RelatesTo>uuid:eaa98952-3188-458f-b265-b03ace115f20</a:RelatesTo>
        <s:NotUnderstood qname="wsman:ResourceUri"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" />
    </s:Header>
    <s:Body>
        <s:Fault>
            <s:Code>
                <s:Value>s:Unknown</s:Value>
            </s:Code>
        </s:Fault>
    </s:Body>
</s:Envelope>"""

    protocol_fake.transport.send_message = lambda m: raise_exc(WinRMTransportError("http", 501, xml_text))

    with pytest.raises(WSManFaultError, match="no error message in fault") as exc:
        protocol_fake.open_shell()

    assert isinstance(exc.value, WSManFaultError)
    assert exc.value.code == 501
    assert exc.value.response == xml_text
    assert exc.value.fault_code == "s:Unknown"
    assert exc.value.fault_subcode is None
    assert exc.value.wsman_fault_code is None
    assert exc.value.wmierror_code is None


def test_wsman_fault_known_fault(protocol_fake):
    xml_text = r"""<s:Envelope xml:lang="en-US"
    xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing"
    xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer"
    xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing"
    xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
    xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"
    xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <a:Action>http://schemas.dmtf.org/wbem/wsman/1/wsman/fault</a:Action>
        <a:MessageID>uuid:D7C4A9B1-9A18-4048-B346-248D62A6078D</a:MessageID>
        <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
        <a:RelatesTo>uuid:7340FE92-C302-42E5-A337-1918908654F8</a:RelatesTo>
    </s:Header>
    <s:Body>
        <s:Fault>
            <s:Code>
                <s:Value>s:Receiver</s:Value>
                <s:Subcode>
                    <s:Value>w:TimedOut</s:Value>
                </s:Subcode>
            </s:Code>
            <s:Reason>
                <s:Text xml:lang="en-US">The WS-Management service cannot complete the operation within the time specified in OperationTimeout.  </s:Text>
            </s:Reason>
            <s:Detail>
                <f:WSManFault
                    xmlns:f="http://schemas.microsoft.com/wbem/wsman/1/wsmanfault" Code="2150858793" Machine="server2022.domain.test">
                    <f:Message>The WS-Management service cannot complete the operation within the time specified in OperationTimeout.  </f:Message>
                </f:WSManFault>
            </s:Detail>
        </s:Fault>
    </s:Body>
</s:Envelope>"""

    protocol_fake.transport.send_message = lambda m: raise_exc(WinRMTransportError("http", 500, xml_text))

    with pytest.raises(WinRMOperationTimeoutError):
        protocol_fake.open_shell()


def test_wsman_fault_with_wsmanfault(protocol_fake):
    xml_text = r"""<s:Envelope xml:lang="en-US"
    xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing"
    xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer"
    xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing"
    xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
    xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"
    xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <a:Action>http://schemas.dmtf.org/wbem/wsman/1/wsman/fault</a:Action>
        <a:MessageID>uuid:EE71C444-1658-4B3F-916D-54CE43B68BC9</a:MessageID>
        <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
        <a:RelatesTo>uuid.761ca906-0bf0-41bb-a9d9-4cbbca986aeb</a:RelatesTo>
    </s:Header>
    <s:Body>
        <s:Fault>
            <s:Code>
                <s:Value>s:Sender</s:Value>
                <s:Subcode>
                    <s:Value>w:SchemaValidationError</s:Value>
                </s:Subcode>
            </s:Code>
            <s:Reason>
                <s:Text xml:lang="">Reason text.</s:Text>
            </s:Reason>
            <s:Detail>
                <f:WSManFault
                    xmlns:f="http://schemas.microsoft.com/wbem/wsman/1/wsmanfault" Code="2150858817" Machine="SERVER2008.domain.local">
                    <f:Message>Detail message.</f:Message>
                </f:WSManFault>
            </s:Detail>
        </s:Fault>
    </s:Body>
</s:Envelope>"""

    protocol_fake.transport.send_message = lambda m: raise_exc(WinRMTransportError("http", 500, xml_text))

    with pytest.raises(WSManFaultError, match="Reason text\\.") as exc:
        protocol_fake.open_shell()

    assert isinstance(exc.value, WSManFaultError)
    assert exc.value.code == 500
    assert exc.value.response == xml_text
    assert exc.value.fault_code == "s:Sender"
    assert exc.value.fault_subcode == "w:SchemaValidationError"
    assert exc.value.wsman_fault_code == 0x80338041
    assert exc.value.wmierror_code is None


def test_wsman_fault_wmi_error_detail(protocol_fake):
    xml_text = r"""<s:Envelope xml:lang="en-US"
    xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing"
    xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer"
    xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing"
    xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
    xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"
    xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <a:Action>http://schemas.dmtf.org/wbem/wsman/1/wsman/fault</a:Action>
        <a:MessageID>uuid:A832545B-9F5C-46AA-BB6A-5E4270D5E530</a:MessageID>
        <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
    </s:Header>
    <s:Body>
        <s:Fault>
            <s:Code>
                <s:Value>s:Receiver</s:Value>
                <s:Subcode>
                    <s:Value>w:InternalError</s:Value>
                </s:Subcode>
            </s:Code>
            <s:Reason>
                <s:Text xml:lang="en-US">Reason text. </s:Text>
            </s:Reason>
            <s:Detail>
                <p:MSFT_WmiError b:IsCIM_Error="true"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xmlns:b="http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd"
                    xmlns:p="http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/MSFT_WmiError"
                    xmlns:cim="http://schemas.dmtf.org/wbem/wscim/1/common" xsi:type="p:MSFT_WmiError_Type">
                    <p:CIMStatusCode xsi:type="cim:cimUnsignedInt">27</p:CIMStatusCode>
                    <p:CIMStatusCodeDescription xsi:type="cim:cimString" xsi:nil="true" />
                    <p:ErrorSource xsi:type="cim:cimString" xsi:nil="true" />
                    <p:ErrorSourceFormat xsi:type="cim:cimUnsignedShort">0</p:ErrorSourceFormat>
                    <p:ErrorType xsi:type="cim:cimUnsignedShort">0</p:ErrorType>
                    <p:Message xsi:type="cim:cimString">WMI Message. </p:Message>
                    <p:MessageID xsi:type="cim:cimString">HRESULT 0x803381a6</p:MessageID>
                    <p:OtherErrorSourceFormat xsi:type="cim:cimString" xsi:nil="true" />
                    <p:OtherErrorType xsi:type="cim:cimString" xsi:nil="true" />
                    <p:OwningEntity xsi:type="cim:cimString" xsi:nil="true" />
                    <p:PerceivedSeverity xsi:type="cim:cimUnsignedShort">0</p:PerceivedSeverity>
                    <p:ProbableCause xsi:type="cim:cimUnsignedShort">0</p:ProbableCause>
                    <p:ProbableCauseDescription xsi:type="cim:cimString" xsi:nil="true" />
                    <p:error_Category xsi:type="cim:cimUnsignedInt">30</p:error_Category>
                    <p:error_Code xsi:type="cim:cimUnsignedInt">2150859174</p:error_Code>
                    <p:error_Type xsi:type="cim:cimString">HRESULT</p:error_Type>
                    <p:error_WindowsErrorMessage xsi:type="cim:cimString">Windows Error message. </p:error_WindowsErrorMessage>
                </p:MSFT_WmiError>
            </s:Detail>
        </s:Fault>
    </s:Body>
</s:Envelope>"""

    protocol_fake.transport.send_message = lambda m: raise_exc(WinRMTransportError("http", 500, xml_text))

    with pytest.raises(WSManFaultError, match="Reason text\\.") as exc:
        protocol_fake.open_shell()

    assert isinstance(exc.value, WSManFaultError)
    assert exc.value.code == 500
    assert exc.value.response == xml_text
    assert exc.value.fault_code == "s:Receiver"
    assert exc.value.fault_subcode == "w:InternalError"
    assert exc.value.wsman_fault_code is None
    assert exc.value.wmierror_code == 0x803381A6



from __future__ import annotations


class WinRMError(Exception):
    """ "Generic WinRM error"""

    code = 500


class WSManFaultError(WinRMError):
    """WSMan Fault Error.

    Exception that is raised when receiving a WSMan fault message. It
    contains the raw response as well as the fault details parsed from the
    response.

    The wsman_fault_code is returned by the Microsoft WSMan server rather than
    the WSMan protocol error code strings. The wmierror_code can contain more
    fatal service error codes returned as a MSFT_WmiError object, for example
    quota violations.

    @param int code: The HTTP status code of the response.
    @param str message: The error message.
    @param str response: The raw WSMan response text.
    @param str reason: The WSMan fault reason.
    @param string fault_code: The WSMan fault code.
    @param string fault_subcode: The WSMan fault subcode.
    @param int wsman_fault_code: The MS WSManFault specific code.
    @param int wmierror_code: The MS WMI error code.
    """

    def __init__(
        self,
        code: int,
        message: str,
        response: str,
        reason: str,
        fault_code: str | None = None,
        fault_subcode: str | None = None,
        wsman_fault_code: int | None = None,
        wmierror_code: int | None = None,
    ) -> None:
        self.code = code
        self.response = response
        self.fault_code = fault_code
        self.fault_subcode = fault_subcode
        self.reason = reason
        self.wsman_fault_code = wsman_fault_code
        self.wmierror_code = wmierror_code

        # Using the dict repr is for backwards compatibility.
        fault_data = {
            "transport_message": message,
            "http_status_code": code,
        }
        if wsman_fault_code is not None:
            fault_data["wsmanfault_code"] = wsman_fault_code

        if fault_code is not None:
            fault_data["fault_code"] = fault_code

        if fault_subcode is not None:
            fault_data["fault_subcode"] = fault_subcode

        super().__init__("{0} (extended fault data: {1})".format(reason, fault_data))


class WinRMTransportError(Exception):
    """WinRM errors specific to transport-level problems (unexpected HTTP error codes, etc)"""

    @property
    def protocol(self) -> str:
        return self.args[0]

    @property
    def code(self) -> int:
        return self.args[1]

    @property
    def message(self) -> str:
        return "Bad HTTP response returned from server. Code {0}".format(self.code)

    @property
    def response_text(self) -> str:
        return self.args[2]

    def __str__(self) -> str:
        return self.message


class WinRMOperationTimeoutError(Exception):
    """
    Raised when a WinRM-level operation timeout (not a connection-level timeout) has occurred. This is
    considered a normal error that should be retried transparently by the client when waiting for output from
    a long-running process.
    """

    code = 500


class AuthenticationError(WinRMError):
    """Authorization Error"""

    code = 401


class BasicAuthDisabledError(AuthenticationError):
    message = "WinRM/HTTP Basic authentication is not enabled on remote host"


class InvalidCredentialsError(AuthenticationError):
    pass



# coding=utf-8
import os
import unittest

import mock

from winrm import transport
from winrm.exceptions import InvalidCredentialsError, WinRMError


class TestTransport(unittest.TestCase):
    maxDiff = 2048
    _old_env = None

    def setUp(self):
        super(TestTransport, self).setUp()
        self._old_env = {}
        os.environ.pop("REQUESTS_CA_BUNDLE", None)
        os.environ.pop("TRAVIS_APT_PROXY", None)
        os.environ.pop("CURL_CA_BUNDLE", None)
        os.environ.pop("HTTPS_PROXY", None)
        os.environ.pop("HTTP_PROXY", None)
        os.environ.pop("NO_PROXY", None)
        transport.DISPLAYED_PROXY_WARNING = False
        transport.DISPLAYED_CA_TRUST_WARNING = False

    def tearDown(self):
        super(TestTransport, self).tearDown()
        os.environ.pop("REQUESTS_CA_BUNDLE", None)
        os.environ.pop("TRAVIS_APT_PROXY", None)
        os.environ.pop("CURL_CA_BUNDLE", None)
        os.environ.pop("HTTPS_PROXY", None)
        os.environ.pop("HTTP_PROXY", None)
        os.environ.pop("NO_PROXY", None)

    def test_build_session_cert_validate_default(self):
        t_default = transport.Transport(
            endpoint="https://example.com",
            username="test",
            password="test",
            auth_method="basic",
        )
        t_default.build_session()
        self.assertEqual(True, t_default.session.verify)

    def test_build_session_cert_validate_default_env(self):
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            username="test",
            password="test",
            auth_method="basic",
        )
        t_default.build_session()
        self.assertEqual("path_to_REQUESTS_CA_CERT", t_default.session.verify)

    def test_build_session_cert_validate_1(self):
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
        )
        t_default.build_session()
        self.assertEqual("path_to_REQUESTS_CA_CERT", t_default.session.verify)

    def test_build_session_cert_validate_2(self):
        os.environ["CURL_CA_BUNDLE"] = "path_to_CURL_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
        )
        t_default.build_session()
        self.assertEqual("path_to_CURL_CA_CERT", t_default.session.verify)

    def test_build_session_cert_override_1(self):
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
            ca_trust_path="overridepath",
        )
        t_default.build_session()
        self.assertEqual("overridepath", t_default.session.verify)

    def test_build_session_cert_override_2(self):
        os.environ["CURL_CA_BUNDLE"] = "path_to_CURL_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
            ca_trust_path="overridepath",
        )
        t_default.build_session()
        self.assertEqual("overridepath", t_default.session.verify)

    def test_build_session_cert_override_3(self):
        os.environ["CURL_CA_BUNDLE"] = "path_to_CURL_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
            ca_trust_path=None,
        )
        t_default.build_session()
        self.assertEqual(True, t_default.session.verify)

    def test_build_session_cert_ignore_1(self):
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"
        os.environ["CURL_CA_BUNDLE"] = "path_to_CURL_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="ignore",
            username="test",
            password="test",
            auth_method="basic",
        )

        t_default.build_session()
        self.assertIs(False, t_default.session.verify)

    def test_build_session_cert_ignore_2(self):
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"
        os.environ["CURL_CA_BUNDLE"] = "path_to_CURL_CA_CERT"

        t_default = transport.Transport(
            endpoint="https://example.com", server_cert_validation="ignore", username="test", password="test", auth_method="basic", ca_trust_path="boguspath"
        )

        t_default.build_session()
        self.assertIs(False, t_default.session.verify)

    def test_build_session_proxy_none(self):
        os.environ["HTTP_PROXY"] = "random_proxy"
        os.environ["HTTPS_PROXY"] = "random_proxy_2"

        t_default = transport.Transport(
            endpoint="https://example.com", server_cert_validation="validate", username="test", password="test", auth_method="basic", proxy=None
        )

        t_default.build_session()
        self.assertEqual({"no_proxy": "*"}, t_default.session.proxies)

    def test_build_session_proxy_defined(self):
        t_default = transport.Transport(
            endpoint="https://example.com", server_cert_validation="validate", username="test", password="test", auth_method="basic", proxy="test_proxy"
        )

        t_default.build_session()
        self.assertEqual({"http": "test_proxy", "https": "test_proxy"}, t_default.session.proxies)

    def test_build_session_proxy_defined_and_env(self):
        os.environ["HTTPS_PROXY"] = "random_proxy"

        t_default = transport.Transport(
            endpoint="https://example.com", server_cert_validation="validate", username="test", password="test", auth_method="basic", proxy="test_proxy"
        )

        t_default.build_session()
        self.assertEqual({"http": "test_proxy", "https": "test_proxy"}, t_default.session.proxies)

    def test_build_session_proxy_with_env_https(self):
        os.environ["HTTPS_PROXY"] = "random_proxy"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
        )

        t_default.build_session()
        self.assertEqual({"https": "random_proxy"}, t_default.session.proxies)

    def test_build_session_proxy_with_env_http(self):
        os.environ["HTTP_PROXY"] = "random_proxy"

        t_default = transport.Transport(
            endpoint="https://example.com",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="basic",
        )

        t_default.build_session()
        self.assertEqual({"http": "random_proxy"}, t_default.session.proxies)

    def test_build_session_server_cert_validation_invalid(self):
        with self.assertRaises(WinRMError) as exc:
            transport.Transport(
                endpoint="Endpoint",
                server_cert_validation="invalid_value",
                username="test",
                password="test",
                auth_method="basic",
            )
        self.assertEqual("invalid server_cert_validation mode: invalid_value", str(exc.exception))

    def test_build_session_krb_delegation_as_str(self):
        winrm_transport = transport.Transport(
            endpoint="Endpoint", server_cert_validation="validate", username="test", password="test", auth_method="kerberos", kerberos_delegation="True"
        )
        self.assertTrue(winrm_transport.kerberos_delegation)

    def test_build_session_krb_delegation_as_invalid_str(self):
        with self.assertRaises(ValueError) as exc:
            transport.Transport(
                endpoint="Endpoint",
                server_cert_validation="validate",
                username="test",
                password="test",
                auth_method="kerberos",
                kerberos_delegation="invalid_value",
            )
        self.assertEqual("invalid truth value 'invalid_value'", str(exc.exception))

    def test_build_session_no_username(self):
        with self.assertRaises(InvalidCredentialsError) as exc:
            transport.Transport(
                endpoint="Endpoint",
                server_cert_validation="validate",
                password="test",
                auth_method="basic",
            )
        self.assertEqual("auth method basic requires a username", str(exc.exception))

    def test_build_session_no_password(self):
        with self.assertRaises(InvalidCredentialsError) as exc:
            transport.Transport(
                endpoint="Endpoint",
                server_cert_validation="validate",
                username="test",
                auth_method="basic",
            )
        self.assertEqual("auth method basic requires a password", str(exc.exception))

    def test_build_session_invalid_auth(self):
        winrm_transport = transport.Transport(
            endpoint="Endpoint",
            server_cert_validation="validate",
            username="test",
            password="test",
            auth_method="invalid_value",
        )

        with self.assertRaises(WinRMError) as exc:
            winrm_transport.build_session()
        self.assertEqual("unsupported auth method: invalid_value", str(exc.exception))

    def test_build_session_invalid_encryption(self):

        with self.assertRaises(WinRMError) as exc:
            transport.Transport(
                endpoint="Endpoint",
                server_cert_validation="validate",
                username="test",
                password="test",
                auth_method="basic",
                message_encryption="invalid_value",
            )
        self.assertEqual("invalid message_encryption arg: invalid_value. Should be 'auto', 'always', or 'never'", str(exc.exception))

    @mock.patch("requests.Session")
    def test_close_session(self, mock_session):
        t_default = transport.Transport(
            endpoint="Endpoint",
            server_cert_validation="ignore",
            username="test",
            password="test",
            auth_method="basic",
        )
        t_default.build_session()
        t_default.close_session()
        mock_session.return_value.close.assert_called_once_with()
        self.assertIsNone(t_default.session)

    @mock.patch("requests.Session")
    def test_close_session_not_built(self, mock_session):
        t_default = transport.Transport(
            endpoint="Endpoint",
            server_cert_validation="ignore",
            username="test",
            password="test",
            auth_method="basic",
        )
        t_default.close_session()
        self.assertFalse(mock_session.return_value.close.called)
        self.assertIsNone(t_default.session)



from __future__ import annotations

import os
import typing as t
import warnings

import requests
import requests.auth

from winrm.encryption import Encryption
from winrm.exceptions import InvalidCredentialsError, WinRMError, WinRMTransportError

DISPLAYED_PROXY_WARNING = False
DISPLAYED_CA_TRUST_WARNING = False


HAVE_KERBEROS = False
try:
    from winrm.vendor.requests_kerberos import REQUIRED, HTTPKerberosAuth

    HAVE_KERBEROS = True
except ImportError:
    pass

HAVE_NTLM = False
try:
    from requests_ntlm import HttpNtlmAuth

    HAVE_NTLM = True
except ImportError as ie:
    pass

HAVE_CREDSSP = False
try:
    from requests_credssp import HttpCredSSPAuth

    HAVE_CREDSSP = True
except ImportError as ie:
    pass

__all__ = ["Transport"]


def strtobool(value: str) -> bool:
    value = value.lower()
    if value in ("true", "t", "yes", "y", "on", "1"):
        return True

    elif value in ("false", "f", "no", "n", "off", "0"):
        return False

    else:
        raise ValueError("invalid truth value '%s'" % value)


class UnsupportedAuthArgument(Warning):
    pass


class Transport(object):
    def __init__(
        self,
        endpoint: str,
        username: str | None = None,
        password: str | None = None,
        realm: None = None,
        service: str | None = None,
        keytab: None = None,
        ca_trust_path: t.Literal["legacy_requests"] | str = "legacy_requests",
        cert_pem: str | None = None,
        cert_key_pem: str | None = None,
        read_timeout_sec: int | None = None,
        server_cert_validation: t.Literal["validate", "ignore"] | None = "validate",
        kerberos_delegation: bool | str = False,
        kerberos_hostname_override: str | None = None,
        auth_method: t.Literal["auto", "basic", "certificate", "ntlm", "kerberos", "credssp", "plaintext", "ssl"] = "auto",
        message_encryption: t.Literal["auto", "always", "never"] = "auto",
        credssp_disable_tlsv1_2: bool = False,
        credssp_auth_mechanism: t.Literal["auto", "ntlm", "kerberos"] = "auto",
        credssp_minimum_version: int = 2,
        send_cbt: bool = True,
        proxy: t.Literal["legacy_requests"] | str | None = "legacy_requests",
    ) -> None:
        self.endpoint = endpoint
        self.username = username
        self.password = password
        self.realm = realm
        self.service = service
        self.keytab = keytab
        self.ca_trust_path = ca_trust_path
        self.cert_pem = cert_pem
        self.cert_key_pem = cert_key_pem
        self.read_timeout_sec = read_timeout_sec
        self.server_cert_validation = server_cert_validation
        self.kerberos_hostname_override = kerberos_hostname_override
        self.message_encryption = message_encryption
        self.credssp_disable_tlsv1_2 = credssp_disable_tlsv1_2
        self.credssp_auth_mechanism = credssp_auth_mechanism
        self.credssp_minimum_version = credssp_minimum_version
        self.send_cbt = send_cbt
        self.proxy = proxy

        if self.server_cert_validation not in [None, "validate", "ignore"]:
            raise WinRMError("invalid server_cert_validation mode: %s" % self.server_cert_validation)

        # defensively parse this to a bool
        if isinstance(kerberos_delegation, bool):
            self.kerberos_delegation = kerberos_delegation
        else:
            self.kerberos_delegation = bool(strtobool(str(kerberos_delegation)))

        self.auth_method = auth_method
        self.default_headers = {
            "Content-Type": "application/soap+xml;charset=UTF-8",
            "User-Agent": "Python WinRM client",
        }

        # try to suppress user-unfriendly warnings from requests' vendored urllib3
        try:
            from requests.packages.urllib3.exceptions import InsecurePlatformWarning

            warnings.simplefilter("ignore", category=InsecurePlatformWarning)
        except Exception:
            pass  # oh well, we tried...

        try:
            from requests.packages.urllib3.exceptions import SNIMissingWarning

            warnings.simplefilter("ignore", category=SNIMissingWarning)
        except Exception:
            pass  # oh well, we tried...

        # if we're explicitly ignoring validation, try to suppress InsecureRequestWarning, since the user opted-in
        if self.server_cert_validation == "ignore":
            try:
                from requests.packages.urllib3.exceptions import InsecureRequestWarning

                warnings.simplefilter("ignore", category=InsecureRequestWarning)
            except Exception:
                pass  # oh well, we tried...

            try:
                from urllib3.exceptions import InsecureRequestWarning

                warnings.simplefilter("ignore", category=InsecureRequestWarning)
            except Exception:
                pass  # oh well, we tried...

        # validate credential requirements for various auth types
        if self.auth_method != "kerberos":
            if self.auth_method == "certificate" or (self.auth_method == "ssl" and (self.cert_pem or self.cert_key_pem)):
                if not self.cert_pem or not self.cert_key_pem:
                    raise InvalidCredentialsError("both cert_pem and cert_key_pem must be specified for cert auth")
                if not os.path.exists(self.cert_pem):
                    raise InvalidCredentialsError("cert_pem file not found (%s)" % self.cert_pem)
                if not os.path.exists(self.cert_key_pem):
                    raise InvalidCredentialsError("cert_key_pem file not found (%s)" % self.cert_key_pem)

            else:
                if not self.username:
                    raise InvalidCredentialsError("auth method %s requires a username" % self.auth_method)
                if self.password is None:
                    raise InvalidCredentialsError("auth method %s requires a password" % self.auth_method)

        self.session: requests.Session | None = None

        # Used for encrypting messages
        self.encryption: Encryption | None = None  # The Pywinrm Encryption class used to encrypt/decrypt messages
        if self.message_encryption not in ["auto", "always", "never"]:
            raise WinRMError("invalid message_encryption arg: %s. Should be 'auto', 'always', or 'never'" % self.message_encryption)

    def build_session(self) -> requests.Session:
        if self.session:
            return self.session

        session = requests.Session()
        proxies = dict()

        if self.proxy is None:
            proxies["no_proxy"] = "*"
        elif self.proxy != "legacy_requests":
            # If there was a proxy specified then use it
            proxies["http"] = self.proxy
            proxies["https"] = self.proxy

        # Merge proxy environment variables
        settings = session.merge_environment_settings(url=self.endpoint, proxies=proxies, stream=None, verify=None, cert=None)

        global DISPLAYED_PROXY_WARNING

        # We want to eventually stop reading proxy information from the environment.
        # Also only display the warning once. This method can be called many times during an application's runtime.
        if not DISPLAYED_PROXY_WARNING and self.proxy == "legacy_requests" and ("http" in settings["proxies"] or "https" in settings["proxies"]):
            message = "'pywinrm' will use an environment defined proxy. This feature will be disabled in " "the future, please specify it explicitly."
            if "http" in settings["proxies"]:
                message += " HTTP proxy {proxy} discovered.".format(proxy=settings["proxies"]["http"])
            if "https" in settings["proxies"]:
                message += " HTTPS proxy {proxy} discovered.".format(proxy=settings["proxies"]["https"])

            DISPLAYED_PROXY_WARNING = True
            warnings.warn(message, DeprecationWarning)

        session.proxies = settings["proxies"]

        # specified validation mode takes precedence
        session.verify = self.server_cert_validation == "validate"

        # patch in CA path override if one was specified in init or env
        if session.verify:
            if self.ca_trust_path == "legacy_requests" and settings["verify"] is not None:
                # We will
                session.verify = settings["verify"]

                global DISPLAYED_CA_TRUST_WARNING

                # We want to eventually stop reading proxy information from the environment.
                # Also only display the warning once. This method can be called many times during an application's runtime.
                if not DISPLAYED_CA_TRUST_WARNING and session.verify is not True:
                    message = (
                        "'pywinrm' will use an environment variable defined CA Trust. This feature will be disabled in "
                        "the future, please specify it explicitly."
                    )
                    if os.environ.get("REQUESTS_CA_BUNDLE") is not None:
                        message += " REQUESTS_CA_BUNDLE contains {ca_path}".format(ca_path=os.environ.get("REQUESTS_CA_BUNDLE"))
                    elif os.environ.get("CURL_CA_BUNDLE") is not None:
                        message += " CURL_CA_BUNDLE contains {ca_path}".format(ca_path=os.environ.get("CURL_CA_BUNDLE"))

                    DISPLAYED_CA_TRUST_WARNING = True
                    warnings.warn(message, DeprecationWarning)

            elif session.verify and self.ca_trust_path is not None:
                # session.verify can be either a bool or path to a CA store; prefer passed-in value over env if both are present
                session.verify = self.ca_trust_path

        encryption_available = False

        if self.auth_method == "kerberos":
            if not HAVE_KERBEROS:
                raise WinRMError("requested auth method is kerberos, but pykerberos is not installed")

            kerb_auth = session.auth = HTTPKerberosAuth(
                mutual_authentication=REQUIRED,
                delegate=self.kerberos_delegation,
                force_preemptive=True,
                principal=self.username,
                hostname_override=self.kerberos_hostname_override,
                sanitize_mutual_error_response=False,
                service=self.service,
                send_cbt=self.send_cbt,
            )
            encryption_available = hasattr(session.auth, "winrm_encryption_available") and kerb_auth.winrm_encryption_available
        elif self.auth_method in ["certificate", "ssl"]:
            if self.auth_method == "ssl" and not self.cert_pem and not self.cert_key_pem:
                # 'ssl' was overloaded for HTTPS with optional certificate auth,
                # fall back to basic auth if no cert specified
                session.auth = requests.auth.HTTPBasicAuth(
                    username=self.username or "",
                    password=self.password or "",
                )
            else:
                session.cert = (self.cert_pem or "", self.cert_key_pem or "")
                session.headers["Authorization"] = "http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual"
        elif self.auth_method == "ntlm":
            if not HAVE_NTLM:
                raise WinRMError("requested auth method is ntlm, but requests_ntlm is not installed")

            session.auth = HttpNtlmAuth(
                username=self.username,
                password=self.password,
                send_cbt=self.send_cbt,
            )
            # check if requests_ntlm has the session_security attribute available for encryption
            encryption_available = hasattr(session.auth, "session_security")
        # TODO: ssl is not exactly right here- should really be client_cert
        elif self.auth_method in ["basic", "plaintext"]:
            session.auth = requests.auth.HTTPBasicAuth(
                username=self.username or "",
                password=self.password or "",
            )
        elif self.auth_method == "credssp":
            if not HAVE_CREDSSP:
                raise WinRMError("requests auth method is credssp, but requests-credssp is not installed")

            session.auth = HttpCredSSPAuth(
                username=self.username,
                password=self.password,
                disable_tlsv1_2=self.credssp_disable_tlsv1_2,
                auth_mechanism=self.credssp_auth_mechanism,
                minimum_version=self.credssp_minimum_version,
            )
            encryption_available = True
        else:
            raise WinRMError("unsupported auth method: %s" % self.auth_method)

        session.headers.update(self.default_headers)
        self.session = session

        # Will check the current config and see if we need to setup message encryption
        if self.message_encryption == "always" and not encryption_available:
            raise WinRMError("message encryption is set to 'always' but the selected auth method %s does not support it" % self.auth_method)
        elif encryption_available:
            if self.message_encryption == "always":
                self.setup_encryption(session)
            elif self.message_encryption == "auto" and not self.endpoint.lower().startswith("https"):
                self.setup_encryption(session)

        return session

    def setup_encryption(self, session: requests.Session) -> None:
        # Security context doesn't exist, sending blank message to initialise context
        request = requests.Request("POST", self.endpoint, data=None)
        prepared_request = session.prepare_request(request)
        self._send_message_request(session, prepared_request)
        self.encryption = Encryption(session, self.auth_method)

    def close_session(self) -> None:
        if not self.session:
            return
        self.session.close()
        self.session = None

    def send_message(self, message: str | bytes) -> bytes:
        session = self.build_session()

        # urllib3 fails on SSL retries with unicode buffers- must send it a byte string
        # see https://github.com/shazow/urllib3/issues/717
        if isinstance(message, str):
            message = message.encode("utf-8")

        if self.encryption:
            prepared_request = self.encryption.prepare_encrypted_request(session, self.endpoint, message)
        else:
            request = requests.Request("POST", self.endpoint, data=message)
            prepared_request = session.prepare_request(request)

        response = self._send_message_request(session, prepared_request)
        return self._get_message_response_text(response)

    def _send_message_request(self, session: requests.Session, prepared_request: requests.PreparedRequest) -> requests.Response:
        try:
            response = session.send(prepared_request, timeout=self.read_timeout_sec)
            response.raise_for_status()
            return response
        except requests.HTTPError as ex:
            if ex.response.status_code == 401:
                raise InvalidCredentialsError("the specified credentials were rejected by the server")
            if ex.response.content:
                response_text = self._get_message_response_text(ex.response)
            else:
                response_text = b""

            raise WinRMTransportError("http", ex.response.status_code, response_text.decode())

    def _get_message_response_text(self, response: requests.Response) -> bytes:
        if self.encryption:
            response_text = self.encryption.parse_encrypted_response(response)
        else:
            response_text = response.content
        return response_text



import pytest

from winrm.protocol import Protocol


@pytest.mark.parametrize("func_name", ["build_wsman_header", "_get_soap_header"])
def test_build_wsman_header(func_name, protocol_fake):
    func = getattr(protocol_fake, func_name)
    actual = func("my action", "resource uri", "shell id", "message id")

    assert actual["env:Header"]["a:Action"]["#text"] == "my action"
    assert actual["env:Header"]["w:ResourceURI"]["#text"] == "resource uri"
    assert actual["env:Header"]["a:MessageID"] == "uuid:message id"
    assert actual["env:Header"]["w:SelectorSet"]["w:Selector"]["#text"] == "shell id"


def test_open_shell_and_close_shell(protocol_fake):
    shell_id = protocol_fake.open_shell()
    assert shell_id == "11111111-1111-1111-1111-111111111113"

    protocol_fake.close_shell(shell_id, close_session=True)


def test_run_command_with_arguments_and_cleanup_command(protocol_fake):
    shell_id = protocol_fake.open_shell()
    command_id = protocol_fake.run_command(shell_id, "ipconfig", ["/all"])
    assert command_id == "11111111-1111-1111-1111-111111111114"

    protocol_fake.cleanup_command(shell_id, command_id)
    protocol_fake.close_shell(shell_id)


def test_run_command_without_arguments_and_cleanup_command(protocol_fake):
    shell_id = protocol_fake.open_shell()
    command_id = protocol_fake.run_command(shell_id, "hostname")
    assert command_id == "11111111-1111-1111-1111-111111111114"

    protocol_fake.cleanup_command(shell_id, command_id)
    protocol_fake.close_shell(shell_id)


def test_get_command_output(protocol_fake):
    shell_id = protocol_fake.open_shell()
    command_id = protocol_fake.run_command(shell_id, "ipconfig", ["/all"])
    std_out, std_err, status_code = protocol_fake.get_command_output(shell_id, command_id)
    assert status_code == 0
    assert b"Windows IP Configuration" in std_out
    assert len(std_err) == 0

    protocol_fake.cleanup_command(shell_id, command_id)
    protocol_fake.close_shell(shell_id)


@pytest.mark.parametrize("func_name", ["get_command_output_raw", "_raw_get_command_output"])
def test_get_command_output_raw(func_name, protocol_fake):
    func = getattr(protocol_fake, func_name)
    shell_id = protocol_fake.open_shell()
    command_id = protocol_fake.run_command(shell_id, "ipconfig", ["/all"])

    std_out, std_err, status_code, done = func(shell_id, command_id)
    assert status_code == 0
    assert b"Windows IP Configuration" in std_out
    assert len(std_err) == 0
    assert done is True

    protocol_fake.cleanup_command(shell_id, command_id)
    protocol_fake.close_shell(shell_id)


def test_send_command_input(protocol_fake):
    shell_id = protocol_fake.open_shell()
    command_id = protocol_fake.run_command(shell_id, "cmd")
    protocol_fake.send_command_input(shell_id, command_id, 'echo "hello world" && exit\r\n')
    std_out, std_err, status_code = protocol_fake.get_command_output(shell_id, command_id)
    assert status_code == 0
    assert b"hello world" in std_out
    assert len(std_err) == 0

    protocol_fake.cleanup_command(shell_id, command_id)
    protocol_fake.close_shell(shell_id)


def test_set_timeout_as_sec():
    protocol = Protocol("endpoint", username="username", password="password", read_timeout_sec="30", operation_timeout_sec="29")
    assert protocol.read_timeout_sec == 30
    assert protocol.operation_timeout_sec == 29


def test_fail_set_read_timeout_as_sec():
    with pytest.raises(ValueError) as exc:
        Protocol("endpoint", username="username", password="password", read_timeout_sec="30a", operation_timeout_sec="29")
    assert str(exc.value) == "failed to parse read_timeout_sec as int: " "invalid literal for int() with base 10: '30a'"


def test_fail_set_operation_timeout_as_sec():
    with pytest.raises(ValueError) as exc:
        Protocol("endpoint", username="username", password="password", read_timeout_sec=30, operation_timeout_sec="29a")
    assert str(exc.value) == "failed to parse operation_timeout_sec as int: " "invalid literal for int() with base 10: '29a'"
